# pylint: disable=missing-module-docstring
# pylint: disable=redefined-outer-name
# pylint: disable=too-few-public-methods
# pylint: disable=missing-class-docstring

from typing import Any
from kaggle_driver.dataset import Input, Target


class DummyInput(Input):
    def __init__(self, value: Any) -> None:
        super().__init__()
        self.value = value


class DummyTarget(Target):
    def __init__(self, value: Any) -> None:
        super().__init__()
        self.value = value



import os
import sys

sys.path.insert(0, os.path.abspath("../examples"))

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.extlinks",
    "sphinx.ext.ifconfig",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx_gallery.gen_gallery",
]
source_suffix = ".rst"
master_doc = "index"
project = "Kaggle Driver"
year = "2023"
author = "Christopher Priebe"
copyright = f"{year}, {author}"
version = release = "0.0.0"

autodoc_mock_imports = ["numpy", "torch"]

class ResetArgv:
    def __repr__(self):
        return 'ResetArgv'

    def __call__(self, sphinx_gallery_conf, script_vars):
        if script_vars["src_file"] == "ground_up_mnist_torch.py":
            return ["-h"]

sphinx_gallery_conf = {
     "examples_dirs": "../examples",
     "gallery_dirs": "_build/auto_examples",
     "reset_argv": ResetArgv(),
}

pygments_style = "trac"
templates_path = ["."]
extlinks = {
    "issue": ("https://github.com/christopherpriebe/kaggle-driver/issues/%s", "#"),
    "pr": ("https://github.com/christopherpriebe/kaggle-driver/pull/%s", "PR #"),
}
# on_rtd is whether we are on readthedocs.org
on_rtd = os.environ.get("READTHEDOCS", None) == "True"

if not on_rtd:  # only set the theme if we are building docs locally
    html_theme = "sphinx_rtd_theme"

html_use_smartypants = True
html_last_updated_fmt = "%b %d, %Y"
html_split_index = False
html_sidebars = {
    "**": ["searchbox.html", "globaltoc.html", "sourcelink.html"],
}
html_short_title = f"{project}-{version}"

napoleon_use_ivar = True
napoleon_use_rtype = False
napoleon_use_param = False



"""A module that contains the KaggleInfo class.
"""
import dataclasses
from typing import Callable


@dataclasses.dataclass
class KaggleInfo:
    """A class that contains information about a Kaggle competition.

    :param competition_name: The name of the competition.
    :type competition_name: str
    :param organize_data_fn: A function that organizes the downloaded dataset
        from Kaggle into the correct folders.

        The function should take in the path to the directory containing the
        downloaded dataset, the path to the directory where the train data
        should be stored, and the path to the directory where the test data
        should be stored.
    :type organize_data_fn: Callable[[str, str, str], None]
    """
    competition_name: str
    organize_data_fn: Callable[[str, str, str], None]



"""
Module that contains the command line app.

Why does this file exist, and why not put this in __main__?

  You might be tempted to import things from __main__ later, but that will cause
  problems: the code will get executed twice:

  - When you run `python -m kaggle_driver` python will execute
    ``__main__.py`` as a script. That means there will not be any
    ``kaggle_driver.__main__`` in ``sys.modules``.
  - When you import __main__ it will get executed again (as a module) because
    there"s no ``kaggle_driver.__main__`` in ``sys.modules``.

  Also see (1) from http://click.pocoo.org/5/setuptools/#setuptools-integration
"""
import argparse
from typing import Optional
from kaggle_driver.dataset import Dataset
from kaggle_driver.directory.model_directory import ModelDirectory
from kaggle_driver.driver.driver import Driver
from kaggle_driver.driver.kaggle_info import KaggleInfo


def _builtin_main() -> None:
    """Entry point for the application script.
    """
    raise NotImplementedError("This entry point for Kaggle Driver is not \
                            implemented yet")


def run(dataset: Dataset, kaggle_info: Optional[KaggleInfo] = None) -> None:
    """Entry point for the application script built by the user.

    :param dataset: The dataset.
    :type dataset: Dataset
    :param kaggle_info: The information about the Kaggle competition. If None,
        then the dataset is assumed to be already downloaded and organized
        into the correct folders.
    :type kaggle_info: Optional[KaggleInfo]
    """
    _user_main(dataset, kaggle_info)


def _user_main(dataset: Dataset, kaggle_info: Optional[KaggleInfo]) -> None:
    models: list[str] = ModelDirectory.keys()

    description: str = "A driver for Kaggle competitions."
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument("--verbose", "-v", action="store_true",
                        help="Print verbose output to stdout.")
    # TODO: Add an argument for logging

    subparsers = parser.add_subparsers(dest="subcommands")

    download_parser = subparsers.add_parser("download",
                                            help="Download the dataset.")

    train_parser = subparsers.add_parser("train", help="Train a model.")
    train_parser.add_argument("model", choices=models,
                              help="The model to train.")
    train_parser.add_argument("--model_config_file", type=str,
                              required=False,
                              help="The path to the model config file.")
    train_parser.add_argument("--train_config_file", type=str,
                              required=False,
                              help="The path to the training config file.")

    test_parser = subparsers.add_parser("test", help="Test a model.")
    test_parser.add_argument("model", choices=models,
                             help="The model to test.")
    test_parser.add_argument("--submission_file", type=str,
                             help="The path to the submission file.")
    test_parser.add_argument("--model_config_file", type=str,
                             required=False,
                             help="The path to the model config file.")
    test_parser.add_argument("--test_config_file", type=str,
                             required=False,
                             help="The path to the testing config file.")

    args: argparse.Namespace = parser.parse_args()
    driver = Driver(dataset, kaggle_info=kaggle_info, verbose=args.verbose)

    if args.subcommands == "download":
        driver.download()
    elif args.subcommands == "train":
        driver.train(args.model, args.model_config_file,
                     args.train_config_file)
    elif args.subcommands == "test":
        driver.test(args.model, args.model_config_file, args.test_config_file,
                    args.submission_file)
    else:
        raise ValueError("Invalid subcommand.")



"""A module for formatting and printing messages to the console.
"""
def print_title(title: str) -> None:
    """Prints a title.

    :param title: The title to print.
    :type title: str
    """
    print(f"\n{title}\n{'-' * len(title)}")


def print_msg(msg: str) -> None:
    """Prints a message.

    :param msg: The message to print.
    :type msg: str
    """
    formatted_msg: str = "| " + msg.replace("\n", "\n| ")
    if formatted_msg.endswith("\n| "):
        formatted_msg = formatted_msg[:-3]
    print(formatted_msg)



"""
Ground-Up MNIST Digit Classification with PyTorch
=================================================

An example of a Kaggle competition driver for the MNIST dataset using PyTorch.

This example aims to demonstrate how to use the Kaggle Driver to build an
environment for a Kaggle competition from the ground-up. Therefore, this
example only builds off of the abstract base classes provided by the Kaggle
Driver and does not use any of the provided helper implementations that
are available in the Kaggle Driver.
"""
import shutil
import os
from collections import OrderedDict
from typing import Optional
import numpy.typing as npt
import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
import kaggle_driver as kd


# ==============================================================================
# Define Input and Target subclasses
# ==============================================================================
class MNISTImage(kd.Input):
    """A class that represents an image in the MNIST dataset.

    :param image: The image.
    :type image: npt.NDArray
    """
    _image: npt.NDArray

    def __init__(self, image: npt.NDArray) -> None:
        super().__init__()
        self._image = image

    @property
    def image(self) -> npt.NDArray:
        """The image.

        :return: The image.
        :rtype: npt.NDArray
        """
        return self._image


class MNISTLabel(kd.Target):
    """A class that represents a label in the MNIST dataset.

    :param label: The label.
    :type label: int
    """
    _label: int

    def __init__(self, label: int) -> None:
        super().__init__()
        self._label = label

    @property
    def label(self) -> int:
        """The label.

        :return: The label.
        :rtype: int
        """
        return self._label


# ==============================================================================
# Define a Dataset subclass using the Input and Target subclasses
# ==============================================================================
class MNISTDataset(kd.Dataset):
    """A class that represents the MNIST dataset.
    """

    def load_train(self) -> OrderedDict[str, tuple[kd.Input, kd.Target]]:
        """Loads the raw training dataset and converts it into an ordered
        dictionary of inputs and targets.

        :return: The ordered dictionary of inputs and targets.
        :rtype: OrderedDict[str, tuple[Input, Target]]
        """
        file_path: str = f"{self.raw_train_dir_path}/train.csv"
        data: npt.NDArray = np.loadtxt(file_path, delimiter=",", skiprows=1)
        inputs: npt.NDArray = data[:, 1:]
        targets: npt.NDArray = data[:, 0]
        train_data: OrderedDict[str, tuple[kd.Input, kd.Target]] = OrderedDict()
        for i, _ in enumerate(inputs):
            train_data[str(i)] = (MNISTImage(inputs[i].reshape(28, 28)),
                                  MNISTLabel(targets[i]))
        return train_data

    def load_test(self) -> OrderedDict[str, kd.Input]:
        """Loads the raw test dataset and converts it into an ordered
        dictionary of inputs.

        :return: The ordered dictionary of inputs.
        :rtype: OrderedDict[str, Input]
        """
        file_path: str = f"{self.raw_test_dir_path}/test.csv"
        data: npt.NDArray = np.loadtxt(file_path, delimiter=",", skiprows=1)
        test_data: OrderedDict[str, kd.Input] = OrderedDict()
        for i, _ in enumerate(data):
            test_data[str(i + 1)] = MNISTImage(data[i].reshape(28, 28))
        return test_data

    def store_predictions(self, predictions_file: str,
                          predictions: dict[str, kd.Target]) -> None:
        """Stores the predictions in a file.

        :param predictions_file: The path to the file to store the predictions
            in.
        :type predictions_file: str
        :param predictions: The predictions to store.
        :type predictions: dict[str, Target]
        """
        with open(predictions_file, "w", encoding="utf-8") as file:
            file.write("ImageId,Label\n")
            for prediction_id, prediction in predictions.items():
                file.write(f"{prediction_id},{prediction.label}\n")


# ==============================================================================
# Define the model, which in this case, is a semi-configurable MLP
# ==============================================================================
@kd.model
class MNISTModel(kd.Model):
    """A configurable MLP model for classification using ReLU
    for all hidden layers and softmax for the output layer.

    :param input_layer_width: The width of the input layer.
    :type input_layer_width: int
    :param output_layer_width: The width of the output layer.
    :type output_layer_width: int
    :param num_hidden_layers: The number of hidden layers.
    :type num_hidden_layers: int
    :param hidden_layer_width: The width of the hidden layers. If
        num_hidden_layers is 0, this parameter is ignored. If num_hidden_layers
        is greater than 0, this parameter must be specified.
    :type hidden_layer_width: Optional[int]
    """
    _input_width: int
    _num_hidden: int
    _hidden_width: int
    _output_width: int
    _model: nn.Sequential

    _DEFAULT_NUM_EPOCHS: int = 5
    _DEFAULT_BATCH_SIZE: int = 64
    _DEFAULT_LEARNING_RATE: float = 0.001
    _DEFAULT_MOMENTUM: float = 0.9

    def __init__(self, input_layer_width: int, output_layer_width: int,
                 num_hidden_layers: int,
                 hidden_layer_width: Optional[int] = None,
                 ) -> None:
        if input_layer_width < 1:
            raise ValueError("input_layer_width must be at least 1")
        if output_layer_width < 1:
            raise ValueError("output_layer_width must be at least 1")
        if num_hidden_layers < 0:
            raise ValueError("num_hidden_layers must be non-negative")
        if (num_hidden_layers > 0) and hidden_layer_width is None:
            raise ValueError("hidden_layer_width must be specified if "
                             "num_hidden_layers is non-zero")
        if hidden_layer_width is not None and (hidden_layer_width < 1):
            raise ValueError("hidden_layer_width must be at least 1")

        if num_hidden_layers == 0:
            name: str = f"MNISTModel_{input_layer_width}_{output_layer_width}"
        else:
            name = (f"MNISTModel_{input_layer_width}_{num_hidden_layers}x"
                    f"{hidden_layer_width}_{output_layer_width}")
        super().__init__(name)

        self._input_width = input_layer_width
        self._num_hidden = num_hidden_layers
        self._hidden_width = hidden_layer_width if num_hidden_layers > 0 else 0
        self._output_width = output_layer_width

        model: list[nn.Module] = []
        model.append(nn.Flatten())
        if num_hidden_layers > 0:
            model.append(nn.Linear(input_layer_width, hidden_layer_width))
            model.append(nn.ReLU())
            for _ in range(num_hidden_layers - 1):
                model.append(nn.Linear(hidden_layer_width, hidden_layer_width))
                model.append(nn.ReLU())
            model.append(nn.Linear(hidden_layer_width, output_layer_width))
        else:
            model.append(nn.Linear(input_layer_width, output_layer_width))
        model.append(nn.Softmax(dim=1))
        self._model = nn.Sequential(*model)

    def train(self, train_data: kd.TrainData,
              train_config: kd.TrainConfig) -> kd.TrainResult:
        """Trains the semi-conigurable MLP.

        :param train_data: The training data.
        :type train_data: TrainData
        :param train_config: The training configuration.
        :type train_config: TrainConfig
        :return: The training results.
        :rtype: TrainResult
        """
        def check_data_point(data_point: tuple[kd.Input, kd.Target]) -> None:
            """Checks that the data point is a MNISTImage and MNISTLabel.

            :param data_point: The data point to check.
            :type data_point: tuple[Input, Target]
            :raises TypeError: If the data point is not a MNISTImage and
                MNISTLabel.
            """
            if not isinstance(data_point[0], MNISTImage):
                raise TypeError("data_point[0] must be a MNISTImage")
            if not isinstance(data_point[1], MNISTLabel):
                raise TypeError("data_point[1] must be a MNISTLabel")

        # Create an instance of the default train result class to track
        # statistics during training.
        res = kd.TrainResult()

        # Check that the training data is valid using a local function.
        for data_point in train_data:
            check_data_point(data_point)

        # Check that the training configuration contains a model path which
        # the model will be saved to after training.
        if "model_path" not in train_config:
            raise ValueError("model_path must be specified in train_config")

        # Assign relevant training configuration parameters to local variables.
        # If a parameter is not specified, use the default value.
        model_path: str = train_config["model_path"]
        num_epochs: int = \
            train_config.get("num_epochs", MNISTModel._DEFAULT_NUM_EPOCHS)
        batch_size: int = \
            train_config.get("batch_size", MNISTModel._DEFAULT_BATCH_SIZE)
        learning_rate: float = \
            train_config.get("learning_rate", MNISTModel._DEFAULT_LEARNING_RATE)
        momentum: float = \
            train_config.get("momentum", MNISTModel._DEFAULT_MOMENTUM)

        trainset = torch.utils.data.TensorDataset(
            torch.tensor([d_p[0].image for d_p in train_data.data_points()],
                         dtype=torch.float32),
            torch.tensor([d_p[1].label for d_p in train_data.data_points()],
                         dtype=torch.long))
        dataloader = torch.utils.data.DataLoader(trainset,
                                                 batch_size=batch_size,
                                                 shuffle=True)

        # Define the loss function an optimizer using PyTorch for training.
        loss_fn = nn.CrossEntropyLoss()
        optimizer = optim.SGD(self._model.parameters(), lr=learning_rate,
                              momentum=momentum)

        # Train the model for the specified number of epochs.
        # Collect the average loss for each epoch and add it to the training
        # results.
        self._model.train()
        res["epoch_average_loss"] = []
        for _ in range(num_epochs):
            epoch_loss: float = 0.0
            for batch in dataloader:
                optimizer.zero_grad()
                outputs = self._model(batch[0])
                loss = loss_fn(outputs, batch[1])
                epoch_loss += loss.item()
                loss.backward()
                optimizer.step()
            res["epoch_average_loss"].append(epoch_loss / len(dataloader))

        # Save the model to the specified path.
        torch.save(self._model.state_dict(), model_path)

        return res

    def test(self, test_data: kd.TestData, test_config: kd.TestConfig) \
        -> tuple[dict[str, kd.Target], kd.TestResult]:
        """Tests the semi-configurable MLP.

        :param test_data: The testing data.
        :type test_data: dict[str, Input]
        :param test_config: The test configuration.
        :type test_config: TestConfig
        :return: The predictions and the testing results.
        :rtype: tuple[dict[str, Target], TestResult]
        """
        def check_data_point(data_point: kd.Input) -> None:
            """Checks that the data point is a MNISTImage.

            :param data_point: The data point to check.
            :type data point: Input
            :raises TypeError: If the data point is not a MNISTImage.
            """
            if not isinstance(data_point, MNISTImage):
                raise TypeError("data_point must be a MNISTImage")

        # Create an instance of the default test result class to track
        # statistics during testing.
        res = kd.TestResult()

        # Check that the testing data is valid using a local function.
        for data_point in test_data:
            check_data_point(data_point)

        # Check that the training configuration contains a model path which
        # the model will be saved to after training.
        if "model_path" not in test_config:
            raise ValueError("model_path must be specified in test_config")

        # Assign relevant training configuration parameters to local variables.
        # If a parameter is not specified, use the default value.
        model_path: str = test_config["model_path"]

        # Load the model from the specified path.
        self._model.load_state_dict(torch.load(model_path))

        # Test the model by generating predictions given the test data.
        self._model.eval()
        predictions: dict[str, kd.Target] = {}
        with torch.no_grad():
            for data_point_id, data_point in zip(test_data.data_point_ids(),
                                                test_data.data_points()):
                inputs: torch.Tensor = torch.tensor(data_point.image,
                                                    dtype=torch.float32)
                outputs: torch.Tensor = self._model(inputs.unsqueeze(0))
                raw_prediction: int = torch.argmax(outputs).item()
                predictions[data_point_id] = MNISTLabel(raw_prediction)

        return predictions, res


# ==============================================================================
# Set the name of the competition on Kaggle
# ==============================================================================
competition_name: str = "digit-recognizer"


# ==============================================================================
# Define the location of the data
# ==============================================================================
curr_dir: str = os.path.dirname(os.path.abspath(__file__))
data_loc_info = kd.DataLocInfo(
    raw_train_dir_path=curr_dir + "/../data/raw/train/mnist",
    raw_test_dir_path=curr_dir + "/../data/raw/test/mnist",
    interim_train_dir_path=curr_dir + "/../data/interim/train/mnist",
    interim_test_dir_path=curr_dir + "/../data/interim/test/mnist",
    processed_train_dir_path=curr_dir + "/..data/processed/train/mnist",
    processed_test_dir_path=curr_dir + "/..data/processed/test/mnist",
)


# ==============================================================================
# Define a function to organize the downloaded data into the correct folders
# ==============================================================================
def organize_mnist_data(unzipped_data_dir: str,
                        raw_train_data_dir: str,
                        raw_test_data_dir: str) -> None:
    """Organizes the MNIST data.

    :param unzipped_data_dir: The directory containing the unzipped data.
    :type unzipped_data_dir: str
    :param raw_train_data_dir: The directory where the raw training data should
        be stored.
    :type raw_train_data_dir: str
    :param raw_test_data_dir: The directory where the raw test data should be
        stored.
    :type raw_test_data_dir: str
    """
    if not os.path.exists(raw_train_data_dir):
        os.makedirs(raw_train_data_dir)
    if not os.path.exists(raw_test_data_dir):
        os.makedirs(raw_test_data_dir)
    train_data_file_suffix: str = "train.csv"
    test_data_file_suffix: str = "test.csv"
    shutil.move(f"{unzipped_data_dir}/{train_data_file_suffix}",
                f"{raw_train_data_dir}/{train_data_file_suffix}")
    shutil.move(f"{unzipped_data_dir}/{test_data_file_suffix}",
                f"{raw_test_data_dir}/{test_data_file_suffix}")


# ==============================================================================
# Set information to be used when interfacing with Kaggle directly
# NOTE: This is optional and only needed if you want to use the Kaggle API
#       Alternatively, you can download the data manually and organize it
#       along with uploading the submission file manually
# ==============================================================================
kaggle_info = kd.KaggleInfo(
    competition_name=competition_name,
    organize_data_fn=organize_mnist_data,
)


# ==============================================================================
# Initialize the dataset and driver and call the driver CLI
# ==============================================================================
dataset = MNISTDataset(data_loc_info)
if __name__ == "__main__":
    kd.run(dataset, kaggle_info=kaggle_info)



"""A module that contains the training data class.
"""
from collections import OrderedDict
from typing import Any, Union
from .data import Data
from .input import Input
from .target import Target


def _check_formatted_data(data: OrderedDict[Any, Any]) -> None:
    for key, value in data:
        if not isinstance(key, str):
            raise TypeError(
                "The keys of the data must be strings."
            )
        if not isinstance(value, tuple):
            raise TypeError(
                "The values of the data must be tuples."
            )
        if len(value) != 2:
            raise ValueError(
                "The tuples in the data must have two elements."
            )
        if not isinstance(value[0], Input):
            raise TypeError(
                "The first element of the tuples in the data must be an Input."
            )
        if not isinstance(value[1], Target):
            raise TypeError(
                "The second element of the tuples in the data must be a Target."
            )


TrainingExample = tuple[Input, Target]


class TrainData(Data):
    """A class that contains the training data.

    :param data: The training data.
    :type data: Union[
    OrderedDict[str, tuple[Input, Target]],
    dict[str, tuple[Input, Target]],
    list[tuple[str, tuple[Input, Target]]],
    tuple[tuple[str, tuple[Input, Target]], ...],
    ]
    """
    def __init__(self, data: Union[
        OrderedDict[str, TrainingExample],
        dict[str, TrainingExample],
        list[tuple[str, TrainingExample]],
        tuple[tuple[str, TrainingExample], ...],
    ]) -> None:
        if isinstance(data, OrderedDict):
            formatted_data = data
        elif isinstance(data, (dict, list, tuple)):
            formatted_data = OrderedDict(data)
        else:
            raise TypeError(
                "The data must be an OrderedDict, dict, list, or tuple."
            )

        _check_formatted_data(formatted_data)
        super().__init__(formatted_data)



"""A module that contains the test data class.
"""
from collections import OrderedDict
from typing import Any, Union
from .data import Data
from .input import Input


def _check_formatted_data(data: OrderedDict[Any, Any]) -> None:
    for key, value in data:
        if not isinstance(key, str):
            raise TypeError(
                "The keys of the data must be strings."
            )
        if not isinstance(value, Input):
            raise TypeError(
                "The values of the data must be tuples."
            )


class TestData(Data):
    """A class that contains the test data.

    :param data: The test data.
    :type data: Union[
    OrderedDict[str, Input],
    dict[str, Input],
    list[tuple[str, Input]],
    tuple[tuple[str, Input], ...],
    ]
    """

    def __init__(self, data: Union[
        OrderedDict[str, Input],
        dict[str, Input],
        list[tuple[str, Input]],
        tuple[tuple[str, Input], ...],
    ]) -> None:
        if isinstance(data, OrderedDict):
            formatted_data = data
        elif isinstance(data, (dict, list, tuple)):
            formatted_data = OrderedDict(data)
        else:
            raise TypeError(
                "The data must be an OrderedDict, dict, list, or tuple."
            )

        _check_formatted_data(formatted_data)
        super().__init__(formatted_data)



# pylint: disable=missing-module-docstring
# pylint: disable=redefined-outer-name
# pylint: disable=too-few-public-methods
# pylint: disable=missing-class-docstring

from typing import Any
import pytest
from kaggle_driver.dataset import Input, Target
from kaggle_driver.directory.model_directory import model, ModelDirectory
from kaggle_driver.model import Model


@pytest.fixture
def dummy_model() -> type:
    """Returns a dummy model.

    :return: The dummy model.
    :rtype: type
    """
    class DummyModel(Model):
        def __init__(self) -> None:
            super().__init__("dummy_model")

        def test(self, test_data: dict[str, Input], *args: Any,  # pylint: disable=unused-argument
                    **kwargs: Any) -> dict[str, Target]:
            return {}
    return DummyModel


def test_model_directory_validates_correct_type(dummy_model) -> None:
    """Tests that the model directory validates a correct value type.
    """
    assert ModelDirectory.is_value_valid(dummy_model)


def test_model_directory_invalidates_incorrect_type() -> None:
    """Tests that the model directory invalidates an incorrect value type.
    """
    assert not ModelDirectory.is_value_valid("dummy_model")


def test_add_dummy_model_to_model_directory(dummy_model) -> None:
    """Tests adding a dummy model to the model directory.
    """
    dummy_model = model(dummy_model)
    assert ModelDirectory.get(dummy_model.__name__) == dummy_model



"""A module that contains the result classes.
"""
import abc


class Result(abc.ABC, dict):
    """The abstract base class for results.
    """


class TrainResult(Result):
    """The result of a training phase.
    """


class TestResult(Result):
    """The result of a testing phase.
    """



"""A module that contains the pretty printer for the kaggle_driver package.
"""
from typing import Any


def pprint_dict(_dict: dict[Any, Any], indent: int = 0) -> str:
    """Pretty prints a dictionary.

    :param _dict: The dictionary.
    :type _dict: dict
    :param indent: The number of spaces to indent the dictionary by. Defaults
        to 0.
    :type indent: int
    :return: The pretty printed dictionary.
    :rtype: str
    """
    res: str = ""
    for key, value in _dict.items():
        res += " " * (indent + 2) + str(key) + ": "
        if isinstance(value, dict):
            res += "\n" + pprint_dict(value, indent + 2)
        else:
            res += str(value) + "\n"
    return res



"""A package that contains classes and functions to manage the dataset.
"""
from .dataset import Dataset
from .data_loc_info import DataLocInfo
from .input import Input
from .target import Target
from .test_data import TestData
from .train_data import TrainData



"""A module that contains the driver class.
"""
import os
import shutil
import zipfile
from typing import Any, Optional
from kaggle_driver.dataset import Dataset
from kaggle_driver.directory import ModelDirectory
from kaggle_driver.model import Model, TestResult, TrainResult
from kaggle_driver.pprint import pprint_dict
from kaggle_driver.utils import make_dir_if_not_exists, yaml_to_dict
from kaggle_driver.verbose_print import print_msg, print_title
from .kaggle_info import KaggleInfo


class Driver:
    """The driver class that orchestrates execution.

    :param dataset: The dataset.
    :type dataset: Dataset
    :param kaggle_info: The information about the Kaggle competition. If None,
        then the dataset is assumed to be already downloaded and organized
        into the correct folders.
    :type kaggle_info: Optional[KaggleInfo]
    :param verbose: Whether to print verbose output to stdout.
    :type verbose: bool
    """
    _YAML_FILE_EXTENSIONS: tuple[str, ...] = (".yaml", ".yml")
    _dataset: Dataset
    _kaggle_info: Optional[KaggleInfo]
    _verbose: bool

    def __init__(self, dataset: Dataset,
                 kaggle_info: Optional[KaggleInfo] = None,
                 verbose: bool = False) -> None:
        self._dataset = dataset
        self._kaggle_info = kaggle_info
        self._verbose = verbose


    def download(self) -> None:
        """Download the dataset from Kaggle.
        """
        if self._kaggle_info is None:
            raise ValueError("Kaggle info must be provided to download "
                                "the dataset.")
        self._download_dataset(self._kaggle_info)

    def _download_dataset(self, kaggle_info: KaggleInfo) -> None:
        import kaggle  # pylint: disable=import-outside-toplevel
        from kaggle import api  # pylint: disable=import-outside-toplevel

        def authenticate(_api: kaggle.KaggleApi) -> None:
            _api.authenticate()

        def download_competition_files(_api: kaggle.KaggleApi,
                                       competition_name: str,
                                       path: str) -> None:
            _api.competition_download_files(competition_name, path=path)

        authenticate(api)
        temp_data_dir_path = f"{os.getcwd()}/.tmp"
        make_dir_if_not_exists(temp_data_dir_path)
        download_competition_files(api, kaggle_info.competition_name,
                                      temp_data_dir_path)
        zip_file_name: str = os.listdir(temp_data_dir_path)[0]
        with zipfile.ZipFile(f"{temp_data_dir_path}/{zip_file_name}", "r") \
                as zip_file:
            zip_file.extractall(temp_data_dir_path)

        try:
            kaggle_info.organize_data_fn(temp_data_dir_path,
                                        self._dataset.raw_train_dir_path,
                                        self._dataset.raw_test_dir_path)
        except Exception as exception:  # pylint: disable=broad-except
            shutil.rmtree(temp_data_dir_path)
            raise exception

        shutil.rmtree(temp_data_dir_path)

    def train(self, model_name: str, model_config_file: Optional[str],
              train_config_file: Optional[str]) -> None:
        """Trains a model.

        :param model_name: The name of the model.
        :type model_name: str
        :param model_config_file: The path to the model configuration file.
        :type model_config_file: Optional[str]
        :param train_config_file: The path to the train configuration file.
        :type train_config_file: Optional[str]
        """
        model_config: dict[str, Any] = self.read_configuration_file(
            model_config_file) if model_config_file is not None else {}
        train_config: dict[str, Any] = self.read_configuration_file(
            train_config_file) if train_config_file is not None else {}
        model_type: type = ModelDirectory.get(model_name)
        model: Model = model_type(**model_config)

        if self._verbose:
            print_title("Training")
            print_msg(f"Model configuration:\n{pprint_dict(model_config)}")
            print_msg(f"Train configuration:\n{pprint_dict(train_config)}")
            print_msg(f"Model: {model.name}")

        train_results: TrainResult = model.train(self._dataset.train,
                                                 train_config)

        if self._verbose:
            print_msg(f"Train results:\n{pprint_dict(train_results)}")

    def test(self, model_name: str, model_config_file: Optional[str],
             test_config_file: Optional[str], submission_file: str) -> None:
        """Tests a model.

        :param model_name: The name of the model.
        :type model_name: str
        :param model_config_file: The path to the model configuration file.
        :type model_config_file: Optional[str]
        :param test_config_file: The path to the test configuration file.
        :type test_config_file: Optional[str]
        :param submission_file: The path to the submission file.
        :type submission_file: str
        """
        model_config: dict[str, Any] = self.read_configuration_file(
            model_config_file) if model_config_file is not None else {}
        test_config: dict[str, Any] = self.read_configuration_file(
            test_config_file) if test_config_file is not None else {}
        model_type: type = ModelDirectory.get(model_name)
        model: Model = model_type(**model_config)

        if self._verbose:
            print_title("Testing")
            print_msg(f"Model configuration:\n{pprint_dict(model_config)}")
            print_msg(f"Test configuration:\n{pprint_dict(test_config)}")
            print_msg(f"Model: {model.name}")

        predictions, test_results = model.test(self._dataset.test, test_config)
        self._dataset.store_predictions(submission_file, predictions)

        if self._verbose:
            print_msg(f"Test results:\n{pprint_dict(test_results)}")
            print_msg(f"Stored predictions in {submission_file}.")

    @staticmethod
    def read_configuration_file(file_path: str) -> dict[str, Any]:
        """Reads a configuration file in varying formats.
        Supported formats: YAML

        :param file_path: The path to the configuration file.
        :type file_path: str
        :return: The configuration.
        :rtype: dict[str, Any]
        """
        if not os.path.exists(file_path):
            raise ValueError(f"Configuration file {file_path} does not exist.")

        if file_path.endswith(Driver._YAML_FILE_EXTENSIONS):
            return yaml_to_dict(file_path)
        else:
            raise ValueError(f"Configuration file {file_path} has an "
                                "unsupported file extension.")



"""A module that contains the configuration classes."""
import abc
from typing import Any


class Config(abc.ABC, dict):
    """The abstract base class for configurations.
    """

    def __init__(self, **kwargs: Any) -> None:
        super().__init__()
        self.update(kwargs)


class TestConfig(Config):
    """The test configuration class.
    """


class TrainConfig(Config):
    """The training configuration class.
    """



"""A module that contains the abstract base class for datasets.
"""
import abc
from collections import OrderedDict
from typing import Optional
from .data_loc_info import DataLocInfo
from .input import Input
from .target import Target
from .test_data import TestData
from .train_data import TrainData


class Dataset(abc.ABC):
    """Abstract base class for datasets.

    :param data_loc_info: Information pertaining to the paths of the dataset
        and further processed versions of that dataset.
    :type data_loc_info: DataLocInfo
    """
    _data_loc_info: DataLocInfo
    _train: Optional[TrainData]
    _test: Optional[TestData]

    def __init__(self, data_loc_info: DataLocInfo) -> None:
        super().__init__()
        self._data_loc_info = data_loc_info
        self._train = None
        self._test = None

    @property
    def raw_train_dir_path(self) -> str:
        """The path to the directory containing the raw training data.

        :return: The path to the directory containing the raw training data.
        :rtype: str
        """
        return self._data_loc_info.raw_train_dir_path

    @property
    def raw_test_dir_path(self) -> str:
        """The path to the directory containing the raw test data.

        :return: The path to the directory containing the raw test data.
        :rtype: str
        """
        return self._data_loc_info.raw_test_dir_path

    @property
    def interim_train_dir_path(self) -> Optional[str]:
        """The path to the directory that will contain the interim training
            data.

        :return: The path to the directory that will contain the interim
            training data.
        :rtype: Optional[str]
        """
        return self._data_loc_info.interim_train_dir_path

    @property
    def interim_test_dir_path(self) -> Optional[str]:
        """The path to the directory that will contain the interim test data.

        :return: The path to the directory that will contain the interim
            test data.
        :rtype: Optional[str]
        """
        return self._data_loc_info.interim_test_dir_path

    @property
    def processed_train_dir_path(self) -> Optional[str]:
        """The path to the directory that will contain the processed training
            data.

        :return: The path to the directory that will contain the processed
            training data.
        :rtype: Optional[str]
        """
        return self._data_loc_info.processed_train_dir_path

    @property
    def processed_test_dir_path(self) -> Optional[str]:
        """The path to the directory that will contain the processed test
            data.

        :return: The path to the directory that will contain the processed
            test data.
        :rtype: Optional[str]
        """
        return self._data_loc_info.processed_test_dir_path

    @property
    def train(self) -> TrainData:
        """The training data.

        :return: The training data.
        :rtype: TrainData
        """
        if self._train is None:
            self._load_train()
        assert self._train is not None, "Something went wrong..."
        return self._train

    @property
    def test(self) -> TestData:
        """The test data.

        :return: The test data.
        :rtype: TestData
        """
        if self._test is None:
            self._load_test()
        assert self._test is not None, "Something went wrong..."
        return self._test

    def _load_train(self) -> None:
        """Loads the training dataset into the internal dictionary of inputs
            and targets.
        """
        self._train = TrainData(self.load_train())

    def _load_test(self) -> None:
        """Loads the test dataset into the internal dictionary of inputs.
        """
        self._test = TestData(self.load_test())

    @abc.abstractmethod
    def load_train(self) -> OrderedDict[str, tuple[Input, Target]]:
        """Loads the raw training dataset and converts it into an ordered
            dictionary of inputs and targets.

        :return: The ordered dictionary of inputs and targets.
        :rtype: OrderedDict[str, tuple[Input, Target]]
        """

    @abc.abstractmethod
    def load_test(self) -> OrderedDict[str, Input]:
        """Loads the raw test dataset and converts it into an ordered
            dictionary of inputs.

        :return: The ordered dictionary of inputs.
        :rtype: OrderedDict[str, Input]
        """

    @abc.abstractmethod
    def store_predictions(self, predictions_file: str,
                          predictions: dict[str, Target]) -> None:
        """Stores the predictions in a file.

        :param predictions_file: The path to the file to store the predictions
            in.
        :type predictions_file: str
        :param predictions: The predictions to store.
        :type predictions: dict[str, Target]
        """



"""Information pertaining to the paths of the dataset and further processed
    versions of that dataset.
"""
import dataclasses
from typing import Optional


@dataclasses.dataclass
class DataLocInfo:
    """Information pertaining to the paths of the dataset and further processed
        versions of that dataset.

    :param raw_train_dir_path: The path to the directory containing
        the raw training data.
    :type raw_train_dir_path: str
    :param raw_test_dir_path: The path to the directory containing
        the raw test data.
    :type raw_test_dir_path: str
    :param interim_train_dir_path: The path to the directory
        that will contain the interim training data.
    :type interim_train_dir_path: Optional[str]
    :default interim_train_dir_path: None
    :param interim_test_dir_path: The path to the directory
        that will contain the interim test data.
    :type interim_test_dir_path: Optional[str]
    :default interim_test_dir_path: None
    :param processed_train_dir_path: The path to the directory
        that will contain the processed training data.
    :type processed_train_dir_path: Optional[str]
    :default processed_train_dir_path: None
    :param processed_test_dir_path: The path to the directory
        that will contain the processed test data.
    :type processed_test_dir_path: Optional[str]
    """

    raw_train_dir_path: str
    raw_test_dir_path: str
    interim_train_dir_path: Optional[str] = None
    interim_test_dir_path: Optional[str] = None
    processed_train_dir_path: Optional[str] = None
    processed_test_dir_path: Optional[str] = None



""" kaggle-driver package
"""
__version__ = "0.0.0"

from .cli import run
from .dataset import Dataset, DataLocInfo, Input, Target, TestData, TrainData
from .directory import model
from .driver import KaggleInfo
from .model import Model, TestConfig, TrainConfig, TestResult, TrainResult



# pylint: disable=missing-module-docstring
# pylint: disable=redefined-outer-name
# pylint: disable=too-few-public-methods
# pylint: disable=missing-class-docstring

import pytest
from kaggle_driver.dataset.data_loc_info import DataLocInfo


def test_initialization_failes_with_missing_mandatory_parameters():
    """Tests that the data location information initialization fails when
    mandatory parameters are missing.
    """
    with pytest.raises(TypeError):
        DataLocInfo("raw/train/path") # pylint: disable=no-value-for-parameter


def test_initialization_with_mandatory_parameters():
    """Tests that the data location information is initialized correctly
    when only mandatory parameters are given.
    """
    data_loc_info = DataLocInfo("raw/train/path", "raw/test/path")
    assert isinstance(data_loc_info, DataLocInfo)
    assert isinstance(data_loc_info.raw_train_dir_path, str)
    assert data_loc_info.raw_train_dir_path == "raw/train/path"
    assert isinstance(data_loc_info.raw_test_dir_path, str)
    assert data_loc_info.raw_test_dir_path == "raw/test/path"
    assert data_loc_info.interim_train_dir_path is None
    assert data_loc_info.interim_test_dir_path is None
    assert data_loc_info.processed_train_dir_path is None
    assert data_loc_info.processed_test_dir_path is None


def test_initialization_with_first_optional_parameter():
    """Tests that the data location information is initialized correctly
    when first optional parameter is given.
    """
    data_loc_info = DataLocInfo(
        "raw/train/path", "raw/test/path",
        interim_train_dir_path="interim/train/path"
    )
    assert isinstance(data_loc_info, DataLocInfo)
    assert isinstance(data_loc_info.raw_train_dir_path, str)
    assert data_loc_info.raw_train_dir_path == "raw/train/path"
    assert isinstance(data_loc_info.raw_test_dir_path, str)
    assert data_loc_info.raw_test_dir_path == "raw/test/path"
    assert isinstance(data_loc_info.interim_train_dir_path, str)
    assert data_loc_info.interim_train_dir_path == "interim/train/path"
    assert data_loc_info.interim_test_dir_path is None
    assert data_loc_info.processed_train_dir_path is None
    assert data_loc_info.processed_test_dir_path is None


def test_initialization_with_second_optional_parameter():
    """Tests that the data location information is initialized correctly
    when second optional parameter is given.
    """
    data_loc_info = DataLocInfo(
        "raw/train/path", "raw/test/path",
        interim_test_dir_path="interim/test/path"
    )
    assert isinstance(data_loc_info, DataLocInfo)
    assert isinstance(data_loc_info.raw_train_dir_path, str)
    assert data_loc_info.raw_train_dir_path == "raw/train/path"
    assert isinstance(data_loc_info.raw_test_dir_path, str)
    assert data_loc_info.raw_test_dir_path == "raw/test/path"
    assert data_loc_info.interim_train_dir_path is None
    assert isinstance(data_loc_info.interim_test_dir_path, str)
    assert data_loc_info.interim_test_dir_path == "interim/test/path"
    assert data_loc_info.processed_train_dir_path is None
    assert data_loc_info.processed_test_dir_path is None


def test_initialization_with_third_optional_parameter():
    """Tests that the data location information is initialized correctly
    when third optional parameter is given.
    """
    data_loc_info = DataLocInfo(
        "raw/train/path", "raw/test/path",
        processed_train_dir_path="processed/train/path"
    )
    assert isinstance(data_loc_info, DataLocInfo)
    assert isinstance(data_loc_info.raw_train_dir_path, str)
    assert data_loc_info.raw_train_dir_path == "raw/train/path"
    assert isinstance(data_loc_info.raw_test_dir_path, str)
    assert data_loc_info.raw_test_dir_path == "raw/test/path"
    assert data_loc_info.interim_train_dir_path is None
    assert data_loc_info.interim_test_dir_path is None
    assert isinstance(data_loc_info.processed_train_dir_path, str)
    assert data_loc_info.processed_train_dir_path == "processed/train/path"
    assert data_loc_info.processed_test_dir_path is None


def test_initialization_with_fourth_optional_parameter():
    """Tests that the data location information is initialized correctly
    when fourth optional parameter is given.
    """
    data_loc_info = DataLocInfo(
        "raw/train/path", "raw/test/path",
        processed_test_dir_path="processed/test/path"
    )
    assert isinstance(data_loc_info, DataLocInfo)
    assert isinstance(data_loc_info.raw_train_dir_path, str)
    assert data_loc_info.raw_train_dir_path == "raw/train/path"
    assert isinstance(data_loc_info.raw_test_dir_path, str)
    assert data_loc_info.raw_test_dir_path == "raw/test/path"
    assert data_loc_info.interim_train_dir_path is None
    assert data_loc_info.interim_test_dir_path is None
    assert data_loc_info.processed_train_dir_path is None
    assert isinstance(data_loc_info.processed_test_dir_path, str)
    assert data_loc_info.processed_test_dir_path == "processed/test/path"


def test_initialization_with_all_parameters():
    """Tests that the data location information is initialized correctly
    when all parameters are given.
    """
    data_loc_info = DataLocInfo(
        "raw/train/path", "raw/test/path",
        "interim/train/path", "interim/test/path",
        "processed/train/path", "processed/test/path"
    )
    assert isinstance(data_loc_info, DataLocInfo)
    assert isinstance(data_loc_info.raw_train_dir_path, str)
    assert data_loc_info.raw_train_dir_path == "raw/train/path"
    assert isinstance(data_loc_info.raw_test_dir_path, str)
    assert data_loc_info.raw_test_dir_path == "raw/test/path"
    assert isinstance(data_loc_info.interim_train_dir_path, str)
    assert data_loc_info.interim_train_dir_path == "interim/train/path"
    assert isinstance(data_loc_info.interim_test_dir_path, str)
    assert data_loc_info.interim_test_dir_path == "interim/test/path"
    assert isinstance(data_loc_info.processed_train_dir_path, str)
    assert data_loc_info.processed_train_dir_path == "processed/train/path"
    assert isinstance(data_loc_info.processed_test_dir_path, str)
    assert data_loc_info.processed_test_dir_path == "processed/test/path"



"""A module containing the abstract base class for data.
"""
import abc
from collections import OrderedDict
from typing import Any, TypeVar


Datum = TypeVar("Datum")


class Data(abc.ABC):
    """An abstract base class for data.

    :param data: The data.
    :type data: OrderedDict[str, Datum]
    """
    _data: OrderedDict[str, Any]

    _iter_index: int

    def __init__(self, data: OrderedDict[str, Datum]) -> None:
        if not isinstance(data, OrderedDict):
            raise TypeError("The data must be an OrderedDict.")
        self._data = data.copy()
        self._iter_index = 0

    def data_points(self) -> list[Datum]:
        """Returns the data points.

        :return: The data points.
        :rtype: list[Datum]
        """
        return list(self._data.values())

    def data_point_ids(self) -> list[str]:
        """Returns the IDs of the data points.

        :return: The IDs of the data points.
        :rtype: list[str]
        """
        return list(self._data.keys())

    def __iter__(self) -> "Data":
        self._iter_index = 0
        return self

    def __next__(self) -> Datum:
        if self._iter_index >= len(self._data):
            raise StopIteration
        else:
            self._iter_index += 1
            return self.data_points()[self._iter_index - 1]



"""A module that defines the abstract base class for all directories.
"""
import abc
from typing import Any


class Directory(abc.ABC):
    """An abstract base class for all directories.
    """
    _dir: dict[str, Any] = {}

    @classmethod
    def get(cls: "Directory", key: str) -> Any:
        """Returns the value for a key.

        :param key: The key.
        :type key: str
        :return: The value.
        :rtype: Any
        """
        return cls._dir[key]

    @classmethod
    def set(cls: "Directory", key: str, value: Any) -> None:
        """Sets the value for a key.

        :param key: The key.
        :type key: str
        :param value: The value.
        :type value: Any
        """
        if not cls.is_value_valid(value):
            raise ValueError(f"Invalid type for value {value}")
        cls._dir[key] = value

    @classmethod
    def contains(cls: "Directory", key: str) -> bool:
        """Returns whether a key is in the directory.

        :param key: The key.
        :type key: str
        :return: Whether the key is in the directory.
        :rtype: bool
        """
        return key in cls._dir

    @classmethod
    def keys(cls: "Directory") -> list[str]:
        """Returns the keys in the directory.

        :return: The keys in the directory.
        :rtype: list[str]
        """
        return list(cls._dir.keys())

    @classmethod
    @abc.abstractmethod
    def is_value_valid(cls: "Directory", value: Any) -> bool:
        """Returns whether a value is valid for this directory.

        :param value: The value to check.
        :type value: Any
        :return: Whether the value is valid.
        :rtype: bool
        """



"""A module that contains the abstract base class for models.
"""
import abc
from kaggle_driver.dataset import Target, TestData, TrainData
from .config import TestConfig, TrainConfig
from .result import TestResult, TrainResult


class Model(abc.ABC):
    """Abstract base class for models.
    """
    _name: str

    def __init__(self, name: str) -> None:
        super().__init__()
        self._name = name

    @property
    def name(self) -> str:
        """Returns the name of the model.

        :return: The name of the model.
        :rtype: str
        """
        return self._name

    def train(self, train_data: TrainData,
              train_config: TrainConfig) -> TrainResult:
        """Trains the model.

        :param train_data: The training data.
        :type train_data: TrainData
        :param train_config: The training configuration.
        :type train_config: TrainConfig
        :return: The training results.
        :rtype: TrainResult
        """

    @abc.abstractmethod
    def test(self, test_data: TestData, test_config: TestConfig) \
        -> tuple[dict[str, Target], TestResult]:
        """Tests the model.

        :param test_data: The testing data.
        :type test_data: dict[str, Input]
        :param test_config: The test configuration.
        :type test_config: TestConfig
        :return: The predictions and the testing results.
        :rtype: tuple[dict[str, Target], TestResult]
        """



"""Utility functions for the kaggle_driver package.
"""
import os
from typing import Any


def make_dir_if_not_exists(path: str) -> None:
    """Makes a directory if it does not exist.

    :param path: The path to the directory.
    :type path: str
    """
    if not os.path.exists(path):
        os.mkdir(path)


def yaml_to_dict(yaml_file_path: str) -> dict[str, Any]:
    """Converts a YAML file to a dictionary.

    :param yaml_file_path: The path to the YAML file.
    :type yaml_file_path: str
    :return: The dictionary representation of the YAML file.
    :rtype: dict
    """
    import yaml  # pylint: disable=import-outside-toplevel
    with open(yaml_file_path, "r", encoding="utf-8") as yaml_file:
        return yaml.load(yaml_file, Loader=yaml.FullLoader)



"""A module that handles registering models.
"""
from typing import Any
from kaggle_driver.model import Model
from .directory import Directory


class ModelDirectory(Directory):
    """A class that handles registering models.
    """
    @classmethod
    def is_value_valid(cls: Directory, value: Any) -> bool:
        """Returns whether a value is valid for this directory.

        :param value: The value to check.
        :type value: Any
        :return: Whether the value is valid.
        :rtype: bool
        """
        return isinstance(value, type) and issubclass(value, Model)


def model(_model: type) -> type:
    """Registers a model.

    :param _model: The model to register.
    :type _model: Model
    :return: The model.
    :rtype: Model
    """
    if not isinstance(_model, type):
        raise TypeError(f"Expected type, got {_model}")
    key: str = _model.__name__
    if ModelDirectory.contains(key):
        raise ValueError(f"Model {key} already registered.")
    ModelDirectory.set(key, _model)
    return _model



# pylint: disable=missing-module-docstring
# pylint: disable=redefined-outer-name
# pylint: disable=too-few-public-methods
# pylint: disable=missing-class-docstring

import pytest
from utils import DummyInput
from kaggle_driver.dataset.input import Input


@pytest.mark.parametrize(
    "value",
    [
        None,
        0,
        "fish",
        [0, 1, 2],
        {"a": 0, "b": 1},
        (2, 3, 5),
        DummyInput("fish")
    ],
)
def test_input_holds_value(value) -> None:
    """Tests that the input holds the correct value.
    """
    dummy_input = DummyInput(value)
    assert isinstance(dummy_input, Input)
    assert id(dummy_input.value) == id(value)
    assert dummy_input.value == value
